/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2022, Kassow Robots
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Kassow Robots nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

#ifndef KR2_KORD_PYREMOTER_H
#define KR2_KORD_PYREMOTER_H

#include <string>
#include <iostream>
#include <map>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/ini_parser.hpp>
#include "boost/assign.hpp"

#include <kr2_program_api/api_v1/bundles/custom_device.h>

#include <kr2/utils/configuration_file.h>

std::map<int, std::string> PYREMOTER_STATUS_SDESC = boost::assign::map_list_of \
(0, "Success.")\
(1, "The internal error. Please, notify the support about this.")\
(2, "Command is failed (create/upload).")\
(3, "Failed to parse the credentials.")\
(4, "Parse host's username failed.")\
(5, "Fail. The client *host* is not defined.")\
(7, "In progress. Try later.")\
(100, "An unknown error occurred.");

/// @breif PyRemoter is an interface for the Transferring any type of file using SSH. \par
/// Possible errors are: \par
/// 1 -- The internal error. \par
/// 2 -- Command is failed (create/upload). \par
/// 3 -- Failed to parse the credentials. \par
/// 4 -- Parse host's username failed. \par
/// 5 -- Fail. The client *host* is not defined.
class PyRemoter {

    public:
        PyRemoter() {};
        PyRemoter(std::string host, std::string username): host(host), username(username) {};

        std::string get_host_username_cmd_end()
        {
            std::string cmd_end = " --host " + this->host + " --username " + this->username;
            return cmd_end;
        }
        
        bool check_handler()
        {
            std::string cmd = this->executor + "-t a -c b --username a --host b --check";
            bool ok = _execute(cmd);

            return ok;
        }
        bool upload(const std::string &src, const std::string &dst) 
        {
            std::string cmd = this->executor + "-t ssh_wrap -c upload " + "--src " + src + " --dst " + dst + this->get_host_username_cmd_end();
            bool ok = _execute(cmd);

            return ok;
        }
        bool create_host_folder(const std::string &path)
        {
            std::string cmd = this->executor + "-t ssh_wrap -c execute -x 'mkdir -p " + path + "' "+ this->get_host_username_cmd_end();
            bool ok = _execute(cmd);

            return ok;    
        }

        // might be a good idea to have it in constructor only...
        std::string get_host()                 { return this->host; }
        void set_host(const std::string &)     { this->host = host; }
        std::string get_username()             { return this->username; }
        void set_username(const std::string &) { this->username = username; }

    private:

        // $PY_CLI = /path/to/entry_cli.py
        std::string executor = "runuser -u nucky -- python3 /home/nucky/kr2.root/CBuns/install/KORD/python/entry_cli.py ";
        std::string host     = "host";
        std::string username = "username";
        
        bool _execute(const std::string &cmd)
        {
            bool success = true;
            // std::cout << "Executed --> [" << cmd  << "]" << std::endl;
            auto sys_ret = system(cmd.c_str());
            int code = WEXITSTATUS(sys_ret);//system(cmd.c_str()));
            if (code != 0) success = false;
            return success;
        }   

};

// Please implement your realization based on this function
/// @brief The main handler of uploading logs to the client machine
/// @param \p host 
/// @param \p username 
/// @param \p path -- uploading to the \p path 
CBUN_PCALL uploadLogs(
    const std::string &host, 
    const std::string &username, 
    const std::string &path) 
{
    /* check the path ends with [UNIX]/ or TODO: it should end [Win]\\  */
    //

    PyRemoter h = PyRemoter(host, username); // remote host and remote name
    // std::cout << h.get_host() << " || " << h.get_username() << std::endl;
    

    bool s = h.check_handler();
    // This can fail only in case the $PY_CLI is not defined --> it should refer to [pyremoter/cli.py or pyremoter/entry_cli.py]
    // Currently, the PyRemoter is located into the CBun folder, so it knows exactly where the *entry_cli.py*
    if (!s) { CBUN_PCALL_RET_ERROR(1, PYREMOTER_STATUS_SDESC[1]); }

    bool ok = h.create_host_folder(path);
    if (!ok) { CBUN_PCALL_RET_ERROR(2, "Creating folder ["+path+"] failed. Check *the path* and/or the provided key in *~/.ssh/authorized_keys*"); }

    // 
    std::string gen_path = "~/.kr2/var/log/";
    std::array<std::string, 2> log_fnames{ "latest", "latest-datalogs" }; // ~/.kr2/var/log/

    for (int i = 0; i < log_fnames.size(); i++)
    {
        std::string _log_fname = log_fnames[i];
        ok = h.upload(gen_path+_log_fname, path+"/"+_log_fname); // extra / will not harm for Unix, but Windows will suffer... TODO
        if (!ok) CBUN_PCALL_RET_ERROR(2, "Uploading log ["+_log_fname+"] failed.");
    }

    CBUN_PCALL_RET_OK;
}


CBUN_PCALL uploadDiagnostics(
    const std::string &host, 
    const std::string &username, 
    const std::string &path) 
{
    /* check the path ends with [UNIX]/ or TODO: it should end [Win]\\  */
    //

    PyRemoter h = PyRemoter(host, username); // remote host and remote name
    // std::cout << h.get_host() << " || " << h.get_username() << std::endl;
    

    bool s = h.check_handler();
    // This can fail only in case the $PY_CLI is not defined --> it should refer to [pyremoter/cli.py or pyremoter/entry_cli.py]
    // Currently, the PyRemoter is located into the CBun folder, so it knows exactly where the *entry_cli.py*
    if (!s) { CBUN_PCALL_RET_ERROR(1, PYREMOTER_STATUS_SDESC[1]); }

    bool ok = h.create_host_folder(path);
    if (!ok) { CBUN_PCALL_RET_ERROR(2, "Creating folder ["+path+"] failed. Check *the path* and/or the provided key in *~/.ssh/authorized_keys*"); }

    //
    std::string ffolder = "/tmp/";
    std::string fname = "dashboard.json";
    ok = h.upload(ffolder + fname, path+"/"+fname);
    if (!ok) CBUN_PCALL_RET_ERROR(2, "Uploading diagnostics ["+fname+"] failed.");

    CBUN_PCALL_RET_OK;
}


CBUN_PCALL uploadCalibrationData(
    const std::string &host, 
    const std::string &username, 
    const std::string &path) 
{
    /* check the path ends with [UNIX]/ or TODO: it should end [Win]\\  */
    //

    PyRemoter h = PyRemoter(host, username); // remote host and remote name
    // std::cout << h.get_host() << " || " << h.get_username() << std::endl;
    

    bool s = h.check_handler();
    // This can fail only in case the $PY_CLI is not defined --> it should refer to [pyremoter/cli.py or pyremoter/entry_cli.py]
    // Currently, the PyRemoter is located into the CBun folder, so it knows exactly where the *entry_cli.py*
    if (!s) { CBUN_PCALL_RET_ERROR(1, PYREMOTER_STATUS_SDESC[1]); }

    bool ok = h.create_host_folder(path);
    if (!ok) { CBUN_PCALL_RET_ERROR(2, "Creating folder ["+path+"] failed. Check *the path* and/or the provided key in *~/.ssh/authorized_keys*"); }

    //
    std::string robot_file_path = "/tmp/model.urdf";
    std::string fname = "model.urdf";
    ok = h.upload(robot_file_path, path+"/"+fname);
    if (!ok) CBUN_PCALL_RET_ERROR(2, "Uploading calibration ["+fname+"] failed.");

    //
    // robot_file_path = "/home/nucky/.kr2/var/data/kincal/RUID_1458851260/kr2_robot_a1205_tio_kc.xacro";
    // fname = "kr2robot_model.xacro";
    // ok = h.upload(robot_file_path, path+"/"+fname);
    // if (!ok) CBUN_PCALL_RET_ERROR(2, "Uploading calibration ["+fname+"] failed.");

    CBUN_PCALL_RET_OK;
}


/// @brief SSH Upload using the *INI* file. \par
/// KORD.ini file example: \par
/// [KORD] \par
/// username = alexander \par
/// log_path = ~/kassow_logs_ini_file [optional] \par
/// host     = 192.168.X.Y            [optional] 
CBUN_PCALL baseSshExportLogsIniWithConfig(const ConfifurationFile::PARSED_DATA &a_config, const std::string &a_host = "") {

    std::string username;
    std::string path;
    std::string host;

    // No username --> it must be a fail
    if (!a_config.USERNAME.empty()){
        username = a_config.USERNAME;
    }
    else{
        CLOG_ERR(PYREMOTER_STATUS_SDESC[4]);
        CBUN_PCALL_RET_ERROR(4, PYREMOTER_STATUS_SDESC[4]);
    }

    // Define a default path for logs: "~/kassow_logs";
    path = a_config.LOG_PATH;
    host = a_config.HOST;
    if (host.empty()) {
        host = a_host;
        if (host.empty()){
            CLOG_ERR(PYREMOTER_STATUS_SDESC[5]);
            CBUN_PCALL_RET_ERROR(5, PYREMOTER_STATUS_SDESC[5]);
        }
    }

    return uploadLogs(host, username, path);
}

CBUN_PCALL baseSshExportLogsIni(const std::string &a_fname, const std::string &a_host = "") {

    std::string credentials_path = "/home/nucky/kr2.root/Workspace/Files/";
    std::string credentials = credentials_path + a_fname;

    boost::property_tree::ptree pt;
    try { boost::property_tree::ini_parser::read_ini(credentials, pt); }
    catch (std::exception &e)
    {
        CLOG_ERR(PYREMOTER_STATUS_SDESC[3]);
        CBUN_PCALL_RET_ERROR(3, PYREMOTER_STATUS_SDESC[3]);
    }

    std::string username;
    std::string path;
    std::string host;

    // No username --> it must be a fail
    try { username = pt.get<std::string>("KORD.username"); }
    catch (std::exception &e)
    {
        CLOG_ERR(PYREMOTER_STATUS_SDESC[4]);
        CBUN_PCALL_RET_ERROR(4, PYREMOTER_STATUS_SDESC[4]);
    }
    
    // Define a default path for logs: "~/kassow_logs";
    try { path = pt.get<std::string>("KORD.log_path"); }
    catch (std::exception &e)
    {
        CLOG_ERR("Parse host's log path failed.");
        path = "~/kassow_logs/";
    }

    try { host = pt.get<std::string>("KORD.host"); }
    catch (std::exception &e)
    {
        CLOG_ERR("Parse host failed.");
        host = a_host;
        // if even a_host is empty --> raise Error
        if (host.empty()) {
            CLOG_ERR(PYREMOTER_STATUS_SDESC[5]);
            CBUN_PCALL_RET_ERROR(5, PYREMOTER_STATUS_SDESC[5]);
        }
    }

    return uploadLogs(host, username, path);
}



CBUN_PCALL baseSshExportDiagnosticsIni(const ConfifurationFile::PARSED_DATA &a_config, const std::string &a_host = "") {


    std::string username;
    std::string path;
    std::string host;

    // No username --> it must be a fail
    if (!a_config.USERNAME.empty()){
        username = a_config.USERNAME;
    }
    else{
        CLOG_ERR(PYREMOTER_STATUS_SDESC[4]);
        CBUN_PCALL_RET_ERROR(4, PYREMOTER_STATUS_SDESC[4]);
    }
    
    // Define a default path for diagnostics: "~/kassow_diagnostics";
    path = a_config.DIAGNOSTICS_PATH;

    host = a_config.HOST;
    if (host.empty()) {
        host = a_host;
        if (host.empty()){
            CLOG_ERR(PYREMOTER_STATUS_SDESC[5]);
            CBUN_PCALL_RET_ERROR(5, PYREMOTER_STATUS_SDESC[5]);
        }
    }

    return uploadDiagnostics(host, username, path);
}

CBUN_PCALL baseSshExportCalibrationDataIni(const ConfifurationFile::PARSED_DATA &a_config, const std::string &a_host = "") {

    std::string username;
    std::string path;
    std::string host;

    // No username --> it must be a fail
    if (!a_config.USERNAME.empty()){
        username = a_config.USERNAME;
    }
    else{
        CLOG_ERR(PYREMOTER_STATUS_SDESC[4]);
        CBUN_PCALL_RET_ERROR(4, PYREMOTER_STATUS_SDESC[4]);
    }
    
    // Define a default path for calibration: "~/kassow_calibration";
    path = a_config.CALIBRATION_PATH;

    host = a_config.HOST;
    if (host.empty()) {
        host = a_host;
        if (host.empty()){
            CLOG_ERR(PYREMOTER_STATUS_SDESC[5]);
            CBUN_PCALL_RET_ERROR(5, PYREMOTER_STATUS_SDESC[5]);
        }
    }

    return uploadCalibrationData(host, username, path);
}

#endif // KR2_KORD_PYREMOTER_H
